import numpy as np

from .base import QupentGate


class QupentI(QupentGate):
    """Identity gate for qupents."""

    def __new__(cls):
        obj = super().__new__(cls, np.eye(5))
        return obj


class QupentX(QupentGate):
    """X gate for qupents."""

    def __new__(cls):
        obj = super().__new__(cls, np.array([[0, 0, 0, 0, 1], [1, 0, 0, 0, 0], [0, 1, 0, 0, 0], [0, 0, 1, 0, 0], [0, 0, 0, 1, 0]]))
        return obj


class QupentZ(QupentGate):
    """Z gate for qupents."""

    def __new__(cls):
        d = 5
        omega = np.exp(2j * np.pi / d)
        phases = [omega**k for k in range(d)]
        obj = super().__new__(cls, np.diag(phases))
        return obj


class QupentH(QupentGate):
    """Hadamard gate for qupents."""

    def __new__(cls):
        obj = super().__new__(
            cls,
            np.array(
                [
                    [1, 1, 1, 1, 1],
                    [1, np.exp(2j * np.pi / 5), np.exp(4j * np.pi / 5), np.exp(6j * np.pi / 5), np.exp(8j * np.pi / 5)],
                    [1, np.exp(4j * np.pi / 5), np.exp(8j * np.pi / 5), np.exp(2j * np.pi / 5), np.exp(6j * np.pi / 5)],
                    [1, np.exp(6j * np.pi / 5), np.exp(2j * np.pi / 5), np.exp(8j * np.pi / 5), np.exp(4j * np.pi / 5)],
                    [1, np.exp(8j * np.pi / 5), np.exp(6j * np.pi / 5), np.exp(4j * np.pi / 5), np.exp(2j * np.pi / 5)],
                ]
            )
            / np.sqrt(5),
        )
        return obj


class QupentT(QupentGate):
    """T gate for qupents."""

    def __new__(cls):
        phases = [np.exp(1j * k * np.pi / 10) for k in range(5)]
        obj = super().__new__(cls, np.diag(phases))
        return obj



import qiskit
import numpy as np
import scipy.linalg
import pennylane as qml

from ..base import HermitianOperator
from ..gates.qubit import X, Z, I


class Hamiltonian(HermitianOperator):
    """
    Class representing a Hamiltonian in quantum computing.

    :param input_array: The input array representing the Hamiltonian. Will be converted to a complex numpy array.
    :param hbar: The reduced Planck constant. Default is 1.0 (natural units).
    If you want to use the actual physical value, set hbar to 1.0545718e-34.
    """

    def __new__(cls, input_array, hbar: float = 1.0):
        assert hbar > 0, "The reduced Planck constant must be greater than zero."
        obj = super().__new__(cls, input_array)
        obj.hbar = hbar
        return obj

    @property
    def num_qubits(self) -> int:
        """Number of qubits in the Hamiltonian."""
        return int(np.log2(self.shape[0]))

    def is_multi_qubit(self) -> bool:
        """Check if Hamiltonian involves multiple qubits."""
        return self.num_qubits > 1

    def time_evolution_operator(self, t: float) -> np.ndarray:
        """Time evolution operator U(t) = exp(-iHt/hbar)."""
        return scipy.linalg.expm(-1j * self * t / self.hbar)

    def ground_state_energy(self) -> float:
        """Ground state energy. i.e. the smallest eigenvalue."""
        eigenvalues = self.eigenvalues()
        return eigenvalues[0]

    def ground_state(self) -> np.ndarray:
        """The eigenvector corresponding to the smallest eigenvalue."""
        _, eigenvectors = np.linalg.eigh(self)
        return eigenvectors[:, 0]

    def convert_endianness(self) -> "Hamiltonian":
        """Hamiltonian from big-endian to little-endian and vice versa."""
        perm = np.argsort([int(bin(i)[2:].zfill(self.num_qubits)[::-1], 2) for i in range(2**self.num_qubits)])
        return self[np.ix_(perm, perm)]

    def add_noise(self, noise_strength: float, noise_operator: np.array) -> "Hamiltonian":
        """
        Add noise to the Hamiltonian.
        :param noise_strength: Strength of the noise. Generally in range [0.0001, 0.5].
        :param noise_operator: Operator representing the noise. For example Pauli matrices.
        :return: A new Hamiltonian with noise added.
        """
        assert noise_operator.shape == self.shape, "Noise operator must have the same dimensions as the Hamiltonian."
        return Hamiltonian(self + noise_strength * noise_operator, hbar=self.hbar)

    def to_qiskit(self) -> qiskit.quantum_info.Operator:
        """
        Convert the scikit-q Hamiltonian to a Qiskit Operator object.
        Qiskit using little endian convention, so we permute the order of the qubits.
        :return: Qiskit Operator object
        """
        return qiskit.quantum_info.Operator(self.convert_endianness())

    @staticmethod
    def from_qiskit(operator: qiskit.quantum_info.Operator) -> "Hamiltonian":
        """
        Create a scikit-q Hamiltonian object from a Qiskit Operator object.
        Qiskit using little endian convention, so we permute the order of the qubits.
        :param operator: Qiskit Operator object
        :return: Hamiltonian object
        """
        return Hamiltonian(operator.data).convert_endianness()

    def to_pennylane(self, wires: list[int] | int = None, **kwargs) -> "qml.Hamiltonian":
        """
        Convert the scikit-q Hamiltonian to a PennyLane Hamiltonian.
        :param wires: List of wires to apply the Hamiltonian to
        kwargs are passed to the PennyLane Hamiltonian constructor.
        :return: PennyLane Hamiltonian object
        """
        coefficients = [1.0]
        wires = wires if wires is not None else list(range(self.num_qubits))
        observables = [qml.Hermitian(self, wires=wires)]
        return qml.Hamiltonian(coefficients, observables, **kwargs)

    @staticmethod
    def from_pennylane(hamiltonian: qml.Hamiltonian) -> "Hamiltonian":
        """
        Convert a PennyLane Hamiltonian object to a scikit-q Hamiltonian object.
        :param hamiltonian: PennyLane Hamiltonian object
        :return: Hamiltonian object
        """
        assert len(hamiltonian.ops) == 1, "Only single-term Hamiltonians are supported."
        return Hamiltonian(hamiltonian.ops[0].matrix())


class IsingHamiltonian(Hamiltonian):
    """
    Hamiltonian for the Ising model.
    :param num_qubits: Number of qubits in the system.
    :param J: Interaction strength between qubits.
    :param h: Transverse field strength.
    :param hbar: The reduced Planck constant. Default is 1.0 (natural units).
    """

    def __new__(cls, num_qubits: int, J: float, h: float, hbar: float = 1.0):
        size = 2**num_qubits
        H = np.zeros((size, size), dtype=complex)
        # Interaction term: -J * sum(Z_i Z_{i+1})
        for i in range(num_qubits - 1):
            term = np.eye(1)
            for j in range(num_qubits):
                if j == i or j == i + 1:
                    term = np.kron(term, Z())
                else:
                    term = np.kron(term, I())
            H -= J * term

        # Transverse field term: -h * sum(X_i)
        for i in range(num_qubits):
            term = np.eye(1)
            for j in range(num_qubits):
                if j == i:
                    term = np.kron(term, X())
                else:
                    term = np.kron(term, I())
            H -= h * term

        return super().__new__(cls, H, hbar)


class HeisenbergHamiltonian(Hamiltonian):
    """
    Hamiltonian for the Heisenberg model.
    :param num_qubits: Number of qubits in the system.
    :param J: Interaction strength between qubits.
    :param hbar: The reduced Planck constant. Default is 1.0 (natural units).
    """

    def __new__(cls, num_qubits: int, J: float, hbar: float = 1.0):
        size = 2**num_qubits
        H = np.zeros((size, size), dtype=complex)

        # Interaction term: J * sum(X_i X_{i+1} + Y_i Y_{i+1} + Z_i Z_{i+1})
        for i in range(num_qubits - 1):
            X_i = np.array([[0, 1], [1, 0]])
            Y_i = np.array([[0, -1j], [1j, 0]])
            Z_i = np.array([[1, 0], [0, -1]])
            H += J * (np.kron(np.kron(np.eye(2**i), np.kron(X_i, X_i)), np.eye(2 ** (num_qubits - i - 2))) + np.kron(np.kron(np.eye(2**i), np.kron(Y_i, Y_i)), np.eye(2 ** (num_qubits - i - 2))) + np.kron(np.kron(np.eye(2**i), np.kron(Z_i, Z_i)), np.eye(2 ** (num_qubits - i - 2))))

        return super().__new__(cls, H, hbar)



import qiskit
import numpy as np
import scipy as sp
import pennylane as qml
from qiskit import QuantumCircuit

from ..base import BaseGate


# I, X, Y, Z Pauli matrices
SINGLE_QUBIT_PAULI_MATRICES = [
    np.eye(2, dtype=complex),  # I
    np.array([[0, 1], [1, 0]], dtype=complex),  # X
    np.array([[0, -1j], [1j, 0]], dtype=complex),  # Y
    np.array([[1, 0], [0, -1]], dtype=complex),  # Z
]
# Any two-qubit Pauli matrix can be expressed as a Kronecker product of two single-qubit Pauli matrices
TWO_QUBIT_PAULI_MATRICES = [np.kron(pauli1, pauli2) for pauli1 in SINGLE_QUBIT_PAULI_MATRICES for pauli2 in SINGLE_QUBIT_PAULI_MATRICES]

# X, Y, Z, H and S gates are Clifford gates
SINGLE_QUBIT_CLIFFORD_MATRICES = SINGLE_QUBIT_PAULI_MATRICES + [
    np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2),  # H
    np.array([[1, 0], [0, 1j]], dtype=complex),  # S
]
# Two qubit Clifford gates can be expressed as a Kronecker product of two single-qubit Clifford gates + CNOT and SWAP
TWO_QUBIT_CLIFFORD_MATRICES = [np.kron(clifford1, clifford2) for clifford1 in SINGLE_QUBIT_CLIFFORD_MATRICES for clifford2 in SINGLE_QUBIT_CLIFFORD_MATRICES] + [
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]], dtype=complex),  # CX (CNOT)
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]], dtype=complex),  # CY
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]], dtype=complex),  # CZ
    np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=complex),  # SWAP
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)], [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)]], dtype=complex),  # CH
    np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1j]], dtype=complex),  # CS
]


class QubitGate(BaseGate):
    """
    Base class for Qubit gates.
    A quantum system with 2 basis states (|0>, |1>).
    Models spin-1/2 particles like electrons.
    """

    def __new__(cls, input_array):
        obj = super().__new__(cls, input_array)
        assert obj.is_at_least_nxn(n=2), "Gate must be at least a 2x2 matrix."
        assert obj.is_power_of_n_shape(n=2), "Gate shape must be a power of 2."
        return obj

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits involved in the gate."""
        return int(np.log2(self.shape[0]))

    def is_multi_qubit(self) -> bool:
        """Check if the gate involves multiple qubits."""
        return self.num_qubits > 1

    def is_pauli(self) -> bool:
        """Check if the gate is a Pauli gate."""
        # I, X, Y, Z Pauli matrices
        if self.num_qubits == 1:
            return any(np.allclose(self, pauli) for pauli in SINGLE_QUBIT_PAULI_MATRICES)
        # Combinations of single-qubit Pauli matrices
        elif self.num_qubits == 2:
            return any(np.allclose(self, pauli) for pauli in TWO_QUBIT_PAULI_MATRICES)
        else:
            return NotImplementedError("Pauli check not supported for gates with more than 2 qubits")

    def is_clifford(self) -> bool:
        """Check if the gate is a Clifford gate."""
        # X, Y, Z, H and S
        if self.num_qubits == 1:
            return any(np.allclose(self, clifford) for clifford in SINGLE_QUBIT_CLIFFORD_MATRICES)
        # Combinations of single-qubit Clifford gates + CNOT and SWAP
        elif self.num_qubits == 2:
            return any(np.allclose(self, clifford) for clifford in TWO_QUBIT_CLIFFORD_MATRICES)
        else:
            return NotImplementedError("Clifford check not supported for gates with more than 2 qubits")

    def sqrt(self) -> "CustomQubitGate":
        """Compute the square root of the gate."""
        sqrt_matrix = sp.linalg.sqrtm(self)
        return CustomQubitGate(sqrt_matrix)

    def kron(self, other: "QubitGate") -> "CustomQubitGate":
        """Compute the Kronecker product of two gates."""
        kron_matrix = np.kron(self, other)
        return CustomQubitGate(kron_matrix)

    def convert_endianness(self) -> "QubitGate":
        """Convert a gate matrix from big-endian to little-endian and vice versa."""
        if self.num_qubits == 1:
            return self
        permutation = np.arange(2**self.num_qubits).reshape([2] * self.num_qubits).transpose().flatten()
        return self[permutation][:, permutation]

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        """
        Convert gate to a Qiskit Gate object.
        Qiskit using little endian convention, so we permute the order of the qubits.
        :return: Qiskit UnitaryGate object
        """
        gate_name = self.__class__.__name__
        print(f"No to_qiskit defined for '{gate_name}'. Initializing as UnitaryGate.")
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label=gate_name)

    @staticmethod
    def from_qiskit(gate: qiskit.circuit.gate.Gate) -> "CustomQubitGate":
        """
        Convert a Qiskit Gate to skq CustomGate.
        Qiskit using little endian convention, so we permute the order of the qubits.
        :param gate: Qiskit Gate object
        """
        assert isinstance(gate, qiskit.circuit.gate.Gate), "Input must be a Qiskit Gate object"
        return CustomQubitGate(gate.to_matrix()).convert_endianness()

    def to_pennylane(self, wires: list[int] | int = None) -> qml.QubitUnitary:
        """
        Convert gate to a PennyLane QubitUnitary.
        PennyLane use the big endian convention, so no need to reverse the order of the qubits.
        :param wires: List of wires the gate acts on
        :return: PennyLane QubitUnitary object
        """
        gate_name = self.__class__.__name__
        print(f"No to_pennylane defined for '{gate_name}'. Initializing as QubitUnitary.")
        wires = wires if wires is not None else list(range(self.num_qubits))
        return qml.QubitUnitary(self, wires=wires)

    @staticmethod
    def from_pennylane(gate: qml.operation.Operation) -> "CustomQubitGate":
        """
        Convert a PennyLane Operation to skqq CustomGate.
        PennyLane use the big endian convention, so no need to reverse the order of the qubits.
        :param gate: PennyLane Operation object
        """
        assert isinstance(gate, qml.operation.Operation), "Input must be a PennyLane Operation object"
        return CustomQubitGate(gate.matrix())

    def to_qasm(self, qubits: list[int]) -> str:
        """
        Convert gate to a OpenQASM string.
        More information on OpenQASM (2.0): https://arxiv.org/pdf/1707.03429
        OpenQASM specification: https://openqasm.com/intro.html
        Gates should be part of the standard library.
        OpenQASM 2.0 -> qelib1.inc
        OpenQASM 3 -> stdgates.inc
        :param qubits: List of qubit indices the gate acts on
        :return: OpenQASM string representation of the gate
        String representation should define gate, qubits it acts on and ;.
        Example for Hadamard in 1st qubit -> "h q[0];"
        """
        raise NotImplementedError("Conversion to OpenQASM is not implemented.")

    @staticmethod
    def from_qasm(qasm_string: str) -> "QubitGate":
        """Convert a OpenQASM string to scikit-q gate."""
        raise NotImplementedError("Conversion from OpenQASM is not implemented.")
    
    def draw(self, **kwargs):
        """Draw the gate using a Qiskit QuantumCircuit."""
        qc = QuantumCircuit(self.num_qubits)
        qc.append(self.to_qiskit(), range(self.num_qubits))
        return qc.draw(**kwargs)


class CustomQubitGate(QubitGate):
    """Bespoke Qubit gate."""

    def __new__(cls, input_array):
        obj = super().__new__(cls, input_array)
        return obj



import qiskit
import pytest
import numpy as np
import pennylane as qml

from skq.base import Operator


def test_operator():
    # Pauli X Gate
    gate = Operator([[0, 1], [1, 0]])
    assert gate.dtype == complex, "Gate should be complex"
    assert gate.num_levels() == 2, "Gate should have 2 levels"
    assert gate.is_square(), "Gate should be square"
    assert gate.is_power_of_n_shape(2), "Gate shape should be a power of 2"
    assert gate.is_2d(), "Gate should be 2D"
    assert gate.is_at_least_nxn(1), "Gate should be at least 1x1"
    assert gate.is_at_least_nxn(2), "Gate should be at least 2x2"
    assert gate.is_hermitian(), "X Gate should be Hermitian"
    assert not gate.is_identity(), "X Gate should not be the identity"
    assert gate.is_equal(gate), "Gate should be equal to itself"
    assert gate.is_equal(Operator([[0, 1], [1, 0]])), "is_equal with itself should be True"
    trace = gate.trace()
    assert trace == 0, "Trace of X Gate should be 0"
    assert isinstance(trace, complex), "Trace should be a complex number"
    assert gate.frobenius_norm() == np.sqrt(2), "Frobenius norm of X Gate should be sqrt(2)"
    # Hermitian -> Eigenvalues are real
    np.testing.assert_array_almost_equal(gate.eigenvalues(), [-1.0, 1.0])
    np.testing.assert_array_almost_equal(gate.eigenvectors(), [[-1, 1], [1, 1]] / np.sqrt(2))

    zero_state = np.array([1, 0])
    assert np.allclose(gate(zero_state), np.array([0, 1])), "X Gate should encode the zero state into the one state"
    assert np.allclose(gate.encodes(zero_state), np.array([0, 1])), "X Gate should encode the zero state into the one state"
    assert np.allclose(gate.decodes(np.array([0, 1])), zero_state), "X Gate should decode the one state into the zero state"

    with pytest.raises(NotImplementedError):
        gate.to_qiskit()
    with pytest.raises(NotImplementedError):
        gate.to_pennylane()
    with pytest.raises(NotImplementedError):
        gate.from_qiskit(qiskit_operator=qiskit.circuit.library.XGate())
    with pytest.raises(NotImplementedError):
        gate.from_pennylane(pennylane_operator=qml.PauliX(wires=0))


def test_invalid_operator():
    # Empty array
    with pytest.raises(AssertionError):
        Operator(np.array([[], []]))
    # Not square
    with pytest.raises(AssertionError):
        Operator(np.array([[1, 0, 0], [0, 1, 0]]))
    # Not 2D
    with pytest.raises(AssertionError):
        Operator(np.array([1, 0, 0]))
    # Not power of 2 shape
    with pytest.raises(AssertionError):
        Operator(np.array([[1, 0], [0, 1], [0, 0], [0, 0]]))



import numpy as np

from .base import QutritGate


class QutritI(QutritGate):
    """
    Identity gate for a qutrit.
    [[1, 0, 0]
    [0, 1, 0]
    [0, 0, 1]]
    """

    def __new__(cls):
        return super().__new__(cls, np.eye(3))


class QutritX(QutritGate):
    """
    X gate for a qutrit.
    |0> -> |1>
    |1> -> |2>
    |2> -> |0>
    """

    def __new__(cls):
        return super().__new__(cls, np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]]))


class QutritY(QutritGate):
    """
    Y gate for a qutrit.
    |0> -> -i|1>
    |1> -> -i|2>
    |2> -> -i|0>
    """

    def __new__(cls):
        return super().__new__(cls, np.array([[0, 0, -1j], [-1j, 0, 0], [0, -1j, 0]]))


class QutritZ(QutritGate):
    """
    Z gate for a qutrit.
    |0> -> |0>
    |1> -> exp(2*pi*i/3)|1>
    |2> -> exp(-2*pi*i/3)|2>
    """

    def __new__(cls):
        return super().__new__(cls, np.array([[1, 0, 0], [0, np.exp(2 * np.pi * 1j / 3), 0], [0, 0, np.exp(-2 * np.pi * 1j / 3)]]))


class QutritH(QutritGate):
    """
    Hadamard gate for a qutrit.
    |0> -> (|0> + |1> + |2>)/sqrt(3)
    |1> -> (|0> + e^(2πi/3)|1> + e^(4πi/3)|2>)/sqrt(3)
    |2> -> (|0> + e^(4πi/3)|1> + e^(2πi/3)|2>)/sqrt(3)
    """

    def __new__(cls):
        omega = np.exp(2j * np.pi / 3)
        return super().__new__(cls, np.array([[1, 1, 1], [1, omega, omega**2], [1, omega**2, omega]]) / np.sqrt(3))


class QutritT(QutritGate):
    """
    T gate for a qutrit.
    |0> -> |0>
    |1> -> exp(2*pi*i/9)|1>
    |2> -> exp(-2*pi*i/9)|2>
    """

    def __new__(cls):
        return super().__new__(cls, np.array([[1, 0, 0], [0, np.exp(2 * np.pi * 1j / 9), 0], [0, 0, np.exp(-2 * np.pi * 1j / 9)]]))


class QutritR(QutritGate):
    """R gate for a qutrit (non-Clifford gate)."""

    def __new__(cls):
        return super().__new__(cls, np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]]))


class QutritPhase(QutritGate):
    """
    General phase gate for qutrits.
    Applies phase shifts to the qutrit basis states.
    """

    def __new__(cls, phi_0: float, phi_1: float, phi_2: float):
        return super().__new__(cls, np.array([[np.exp(1j * phi_0), 0, 0], [0, np.exp(1j * phi_1), 0], [0, 0, np.exp(1j * phi_2)]]))


class QutritS(QutritPhase):
    """
    S gate for qutrits.
    |0> -> |0>
    |1> -> exp(2*pi*i/3)|1>
    |2> -> exp(4*pi*i/3)|2>
    """

    def __new__(cls):
        return super().__new__(cls, phi_0=0, phi_1=2 * np.pi / 3, phi_2=4 * np.pi / 3)



import qiskit
import numpy as np
import pennylane as qml

from .base import QubitGate
from .single import X, Y, Z
from ...quantum_info.state import Statevector


class PhaseOracle(QubitGate):
    """
    Phase Oracle as used in Grover's search algorithm.
    :param target_state: The target state to mark.
    target_state is assumed to be in Big-Endian format.
    """

    def __new__(cls, target_state: np.ndarray):
        state = Statevector(target_state)
        n_qubits = state.num_qubits
        identity = np.eye(2**n_qubits)
        oracle_matrix = identity - 2 * np.outer(target_state, target_state.conj())
        return super().__new__(cls, oracle_matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="PhaseOracle")


class GroverDiffusion(QubitGate):
    """
    Grover Diffusion Operator Gate as used in Grover's search algorithm.
    This gate amplifies the amplitude of the marked state.
    :param n_qubits: The number of qubits in the system.
    """

    def __new__(cls, n_qubits: int):
        assert n_qubits >= 1, "GroverDiffusionGate must have at least one qubit."
        # Equal superposition state vector
        size = 2**n_qubits
        equal_superposition = np.ones(size) / np.sqrt(size)
        # Grover diffusion operator: 2|ψ⟩⟨ψ| - I
        diffusion_matrix = 2 * np.outer(equal_superposition, equal_superposition) - np.eye(size)
        return super().__new__(cls, diffusion_matrix)

    def to_qiskit(self) -> qiskit.circuit.library.UnitaryGate:
        # Reverse the order of qubits for Qiskit's little-endian convention
        return qiskit.circuit.library.UnitaryGate(self.convert_endianness(), label="GroverDiffusion")


class CX(QubitGate):
    """
    Controlled-X (CNOT) gate.
    Used to entangle two qubits.
    If the control qubit is |1>, the target qubit is flipped.
    """

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CXGate:
        return qiskit.circuit.library.CXGate()

    def to_pennylane(self, wires: list[int]) -> qml.CNOT:
        assert len(wires) == 2, "PennyLane CX gate requires 2 wires."
        return qml.CNOT(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CX gate requires 2 qubits."
        return f"cx q[{qubits[0]}], q[{qubits[1]}];"


class CY(QubitGate):
    """Controlled-Y gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CYGate:
        return qiskit.circuit.library.CYGate()

    def to_pennylane(self, wires: list[int]) -> qml.CY:
        assert len(wires) == 2, "PennyLane CY gate requires 2 wires."
        return qml.CY(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CY gate requires 2 qubits."
        return f"cy q[{qubits[0]}], q[{qubits[1]}];"


class CZ(QubitGate):
    """Controlled-Z gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])

    def to_qiskit(self) -> qiskit.circuit.library.CZGate:
        return qiskit.circuit.library.CZGate()

    def to_pennylane(self, wires: list[int]) -> qml.CZ:
        assert len(wires) == 2, "PennyLane CZ gate requires 2 wires."
        return qml.CZ(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CZ gate requires 2 qubits."
        return f"cz q[{qubits[0]}], q[{qubits[1]}];"


class CH(QubitGate):
    """Controlled-Hadamard gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1 / np.sqrt(2), 1 / np.sqrt(2)], [0, 0, 1 / np.sqrt(2), -1 / np.sqrt(2)]])

    def to_qiskit(self) -> qiskit.circuit.library.CHGate:
        return qiskit.circuit.library.CHGate()

    def to_pennylane(self, wires: list[int]) -> qml.CH:
        assert len(wires) == 2, "PennyLane CH gate requires 2 wires."
        return qml.CH(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CH gate requires 2 qubits."
        return f"ch q[{qubits[0]}], q[{qubits[1]}];"


class CPhase(QubitGate):
    """General controlled phase shift gate.
    :param phi: The phase shift angle in radians.
    """

    def __new__(cls, phi: float):
        obj = super().__new__(cls, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(1j * phi)]])
        obj.phi = phi
        return obj

    def to_qiskit(self) -> qiskit.circuit.library.CPhaseGate:
        return qiskit.circuit.library.CPhaseGate(self.phi)

    def to_pennylane(self, wires: list[int]) -> qml.CPhase:
        assert len(wires) == 2, "PennyLane CPhase gate requires 2 wires."
        return qml.CPhase(phi=self.phi, wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM CPhase gate requires 2 qubits."
        return f"cp({self.phi}) q[{qubits[0]}], q[{qubits[1]}];"


class CS(CPhase):
    """Controlled-S gate."""

    def __new__(cls):
        phi = np.pi / 2
        return super().__new__(cls, phi=phi)

    def to_qiskit(self) -> qiskit.circuit.library.CSGate:
        return qiskit.circuit.library.CSGate()


class CT(CPhase):
    """Controlled-T gate."""

    def __new__(cls):
        phi = np.pi / 4
        return super().__new__(cls, phi=phi)

    def to_qiskit(self) -> qiskit.circuit.library.TGate:
        return qiskit.circuit.library.TGate().control(1)


class SWAP(QubitGate):
    """Swap gate. Swaps the states of two qubits."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    def to_qiskit(self) -> qiskit.circuit.library.SwapGate:
        return qiskit.circuit.library.SwapGate()

    def to_pennylane(self, wires: list[int]) -> qml.SWAP:
        assert len(wires) == 2, "PennyLane SWAP gate requires 2 wires."
        return qml.SWAP(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 2, "OpenQASM SWAP gate requires 2 qubits."
        return f"swap q[{qubits[0]}], q[{qubits[1]}];"


class CSwap(QubitGate):
    """A controlled-SWAP gate. Also known as the Fredkin gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, 1]])

    def to_qiskit(self) -> qiskit.circuit.library.CSwapGate:
        return qiskit.circuit.library.CSwapGate()

    def to_pennylane(self, wires: list[int]) -> qml.CSWAP:
        return qml.CSWAP(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CSWAP gate requires 3 qubits."
        return f"cswap q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];"


class CCX(QubitGate):
    """A 3-qubit controlled-controlled-X (CCX) gate. Also known as the Toffoli gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, 1], [0, 0, 0, 0, 0, 0, 1, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.CCXGate:
        return qiskit.circuit.library.CCXGate()

    def to_pennylane(self, wires: list[int]) -> qml.Toffoli:
        return qml.Toffoli(wires=wires)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCX (Toffoli) gate requires 3 qubits."
        return f"ccx q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];"


class CCY(QubitGate):
    """A 3-qubit controlled-controlled-Y (CCY) gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 0, -1j], [0, 0, 0, 0, 0, 0, 1j, 0]])

    def to_qiskit(self) -> qiskit.circuit.library.YGate:
        # There is no native CCY gate in Qiskit so we construct it.
        return qiskit.circuit.library.YGate().control(2)

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCY gate requires 3 qubits."
        # Custom OpenQASM implementation
        # sdg is an inverse s gate
        return f"""sdg q[{qubits[2]}];\ncx q[{qubits[1]}], q[{qubits[2]}];\ns q[{qubits[2]}];\ncx q[{qubits[0]}], q[{qubits[2]}];\nsdg q[{qubits[2]}];\ncx q[{qubits[1]}], q[{qubits[2]}];\ns q[{qubits[2]}];\ny q[{qubits[2]}];"""


class CCZ(QubitGate):
    """A 3-qubit controlled-controlled-Z (CCZ) gate."""

    def __new__(cls):
        return super().__new__(cls, [[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, -1]])

    def to_qiskit(self) -> qiskit.circuit.library.CCZGate:
        return qiskit.circuit.library.CCZGate()

    def to_qasm(self, qubits: list[int]) -> str:
        assert len(qubits) == 3, "OpenQASM CCZ gate requires 3 qubits."
        # CCZ = CCX sandwiched between two H gates on last qubit.
        return f"""h q[{qubits[2]}];\nccx q[{qubits[0]}], q[{qubits[1]}], q[{qubits[2]}];\nh q[{qubits[2]}];"""


class MCX(QubitGate):
    """
    Multi controlled-X (MCX) gate.
    :param num_ctrl_qubits: Number of control qubits.
    """

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCX gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = X()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.CXGate | qiskit.circuit.library.CCXGate | qiskit.circuit.library.C3XGate | qiskit.circuit.library.C4XGate | qiskit.circuit.library.MCXGate:
        if self.num_ctrl_qubits == 1:
            return qiskit.circuit.library.CXGate()
        elif self.num_ctrl_qubits == 2:
            return qiskit.circuit.library.CCXGate()
        elif self.num_ctrl_qubits == 3:
            return qiskit.circuit.library.C3XGate()
        elif self.num_ctrl_qubits == 4:
            return qiskit.circuit.library.C4XGate()
        else:
            return qiskit.circuit.library.MCXGate(num_ctrl_qubits=self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CNOT | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CNOT(wires=wires)
        else:
            return super().to_pennylane(wires)


class MCY(QubitGate):
    """Multi controlled-Y (MCY) gate."""

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCY gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = Y()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.YGate:
        return qiskit.circuit.library.YGate().control(self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CY | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CY(wires=wires)
        else:
            return super().to_pennylane(wires)


class MCZ(QubitGate):
    """Multi controlled-Z (MCZ) gate."""

    def __new__(cls, num_ctrl_qubits: int):
        assert num_ctrl_qubits >= 1, "MCZ gate must have at least one control qubit."
        cls.num_ctrl_qubits = num_ctrl_qubits
        levels = 2 ** (num_ctrl_qubits + 1)
        gate = np.identity(levels)
        gate[-2:, -2:] = Z()
        return super().__new__(cls, gate)

    def to_qiskit(self) -> qiskit.circuit.library.ZGate:
        return qiskit.circuit.library.ZGate().control(self.num_ctrl_qubits)

    def to_pennylane(self, wires: list[int]) -> qml.CZ | qml.QubitUnitary:
        if self.num_ctrl_qubits == 1:
            return qml.CZ(wires=wires)
        else:
            return super().to_pennylane(wires)


class CR(QubitGate):
    """Simple Cross-Resonance gate."""

    def __new__(cls, theta: float):
        return super().__new__(cls, [[1, 0, 0, 0], [0, np.cos(theta), 0, -1j * np.sin(theta)], [0, 0, 1, 0], [0, -1j * np.sin(theta), 0, np.cos(theta)]])


class SymmetricECR(QubitGate):
    """Symmetric Echoed Cross-Resonance gate."""

    def __new__(cls, theta: float):
        return super().__new__(cls, CR(theta) @ CR(-theta))


class AsymmetricECR(QubitGate):
    """Asymmetric Echoed Cross-Resonance gate."""

    def __new__(cls, theta1: float, theta2: float):
        return super().__new__(cls, CR(theta1) @ CR(theta2))
